/*!
 * Copyright (c) 2015-2019 Cisco Systems, Inc. See LICENSE file.
 */

import {StatelessWebexPlugin} from '@webex/webex-core';

import {
  MEETINGS,
  _MEETING_LINK_,
  _SIP_URI_
} from '../constants';

import MeetingInfoCollection from './collection';
import MeetingInfoRequest from './request';
import MeetingInfoUtil from './util';

/**
 * @class MeetingInfo
 */
export default class MeetingInfo extends StatelessWebexPlugin {
  namespace = MEETINGS;

  /**
   *
   * @param {Object} attrs
   * @param {Object} options
   */
  constructor(attrs, options) {
    super({}, options);
    /**
     * The meeting info request server interface
     * @instance
     * @type {Object}
     * @private
     * @memberof MeetingInfo
    */
    this.meetingInfoRequest = new MeetingInfoRequest({}, options);
    /**
     * The meeting information collection interface
     * @instance
     * @type {Object}
     * @private
     * @memberof MeetingInfo
    */
    this.meetingInfoCollection = new MeetingInfoCollection();
  }

  /**
   * @param {string} id
   * @returns {object}
   * @public
   * @memberof MeetingInfo
   */
  getMeetingInfo(id) {
    return this.meetingInfoCollection.get(id);
  }

  /**
   * @param {string} id
   * @param {object} info
   * @returns {null}
   * @public
   * @memberof MeetingInfo
   */
  setMeetingInfo(id, info) {
    this.meetingInfoCollection.set(id, info);
  }

  /**
   * Helper to make the actual MeetingInfo request and set the meetingInfo if successful, else reject
   * @param {Object} options
   * @returns {Promise}
   * @private
   * @memberof MeetingInfo
   */
  requestFetchInfo(options) {
    return this.meetingInfoRequest.fetchMeetingInfo(options).then((info) => {
      if (info && info.body) {
        this.setMeetingInfo(info.body.sipMeetingUri || info.body.meetingLink, info.body);
      }

      return info;
    }).catch((error) => Promise.reject(error));
  }

  /**
   * Helper to generate the options for the MeetingInfo request
   * @param {String} destination
   * @param {String} type
   * @returns {Promise}
   * @private
   * @memberof MeetingInfo
   */
  fetchInfoOptions(destination, type) {
    return MeetingInfoUtil.generateOptions({
      destination,
      type,
      webex: this.webex
    });
  }

  /**
   * Fetches meeting info from the server
   * @param {String} destination one of many different types of destinations to look up info for
   * @param {String} [type] to match up with the destination value
   * @returns {Promise} returns a meeting info object
   * @public
   * @memberof MeetingInfo
   */
  fetchMeetingInfo(destination, type = null) {
    return this.fetchInfoOptions(destination, type).then((options) =>
    // fetch meeting info
      this.requestFetchInfo(options).catch((error) => {
      // if it failed the first time as meeting link
        if (options.type === _MEETING_LINK_) {
        // convert the meeting link to sip URI and retry
          return this.requestFetchInfo(this.fetchInfoOptions(MeetingInfoUtil.convertLinkToSip(destination), _SIP_URI_));
        }

        return Promise.reject(error);
      }));
  }
}
